/*
 * Copyright (C) 2011-2012 @unobfuscator All Rights Reserved.
 *
 * Open source licensed under the MIT license
 * 
 */

package groovy.util;

import groovy.util.BuilderSupport;

import org.codehaus.groovy.runtime.StringBufferWriter
import java.util.Iterator;
import java.util.Map;

public class HBuilder extends BuilderSupport {
    private StringBufferWriter out;
    private boolean nospace;
    private boolean scriptCdata = false;
    private boolean scriptCdataDefault = false;
    private int state;
    private boolean nodeIsEmpty = true;
    private Class mdpClass = null;
    private Class pbClass = null;
    
    public HBuilder() {
        def sb = new StringBuffer(1000);
        this.out = new StringBufferWriter( sb);
        ClassLoader classLoader = HBuilder.class.getClassLoader();
        try {
            this.mdpClass = classLoader.loadClass("nineci.markdownplus.MarkdownPlus");
        } catch (ClassNotFoundException e) {
            this.mdpClass = null;
            try {
                this.pbClass = classLoader.loadClass("com.petebevin.markdown.MarkdownProcessor");
            } catch (ClassNotFoundException ee) {
                this.pbClass = null;
            }
        }
    }
    
    public String getOutput() {
        return this.out.toString();
    }
    
    protected void setParent(Object parent, Object child) {
        
    }
    
    protected Object createNode(Object name) {
        toState(1, name);
        nodeIsEmpty = true
        return name;
    }

    protected Object createNode(Object name, Object value) {
        toState(2, name);
        if ( name.length() > 0) {
            switch( name) {
                default:
                    out.print(">");
                    out.print((scriptCdata ? "\n//<![CDATA[\n" : "") + value + (scriptCdata ? "\n//]]>\n" : ""));
            }
        } else {
            out.print(value.toString());
        }
        if ( value == null) {
            nodeIsEmpty = true
        } else {
            nodeIsEmpty = false
        }
        return name;
    }
    
    protected Object createNode(Object name, Map attributes, Object value) {
        toState(1, name);
        for (Iterator iter = attributes.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            out.print(" ");
            print(transformName(entry.getKey().toString()));
            out.print("='");
            print(transformValue(entry.getValue().toString()));
            out.print("'");
        }
        if (value != null) {
            nodeIsEmpty = false;
            if ( name.length() > 0) {
                switch( name) {
                    default:
                        out.print(">" + (scriptCdata ? "\n//<![CDATA[\n" : "") + value + (scriptCdata ? "\n//]]>\n" : "") +"</" + name + ">");
                        break;
                }
            } else {
                out.print(value);
            }
        } else {
            nodeIsEmpty = true
        }
        return name;
    }
    
    protected Object createNode(Object name, Map attributes) {
        //println "createNode, just attributes"
        return createNode(name, attributes, null);
    }
    
    protected void nodeCompleted(Object parent, Object node) {
        //println "nodeCompleted: state ${state}, nodeIsEmpty ${nodeIsEmpty}"
        toState(3, node);
        out.flush();
    }
    
    protected void print(Object node) {
        out.print(node == null ? "null" : node.toString());
    }
    
    protected Object getName(String methodName) {
        scriptCdata = false
        if ( methodName == "text") {
            methodName = "";
        }
        if ( methodName == "script") {
            scriptCdata = scriptCdataDefault;
        }
        return super.getName(transformName(methodName));
    }
    
    protected String transformName(String name) {
        if (name.startsWith("_"))
            name = name.substring(1);
        return name.replace('_', '-');
    }
    
    protected String transformValue(String value) {
        return value.replaceAll("\\'", "&quot;");
    }
    
    private void toState(int next, Object name) {
        switch (state) {
            case 0:
                switch (next) {
                    case 1:
                    case 2:
                        if ( name.length() > 0) {
                            out.print("<");
                            print(name);
                        }
                        break;
                    case 3:
                        throw new Error();
                }
                break;
            case 1:
                switch (next) {
                    case 1:
                    case 2:
                        out.print(">");
                        if (nospace) {
                            nospace = false;
                        } else {
                            out.println();
                        }
                        if ( name.length() > 0) {
                            out.print("<");
                            print(name);
                        }
                        break;
                    case 3:
                        if (nodeIsEmpty) {
                            out.print(" />");
                        }
                        break;
                }
                break;
            case 2:
                switch (next) {
                    case 1:
                    case 2:
                        throw new Error();
                    case 3:
                        if ( name.length() > 0) {
                            out.print("</");
                            print(name);
                            out.print(">");
                        }
                        break;
                }
                break;
            case 3:
                switch (next) {
                    case 1:
                    case 2:
                        if (nospace) {
                            nospace = false;
                        } else {
                            out.println();
                        }
                        if ( name.length() > 0) {
                            out.print("<");
                            print(name);
                        }
                        break;
                    case 3:
                        if (nospace) {
                            nospace = false;
                        } else {
                            out.println();
                        }
                        if ( name.length() > 0) {
                            out.print("</");
                            print(name);
                            out.print(">");
                        }
                        break;
                }
                break;
        }
        state = next;
        nodeIsEmpty = false
    }
    
    def closePreviousTag() {
        if ( state == 1) {
            if ( nodeIsEmpty) {
                out.println( "/>")
            } else {
                out.println( ">")
            }
        }
        state = 3
        nodeIsEmpty = false
    }
    
    // Pseudo tag functions
    
    def cdata ( String value) {
        cdata( value, (Closure)null)
    }
    
    def cdata ( Closure closure) {
        cdata( "", closure)
    }

    def cdata ( String value, Closure closure) {
        closePreviousTag()
        out.print("<![CDATA[${value}")
        if ( closure) {
            nospace = true // to avoid line break(s) being inserted inside CDATA sections
            closure.setDelegate = this
            closure()
        }
        out.print("]]>")
    }

    def setdefaults( def attributes) {
        for (Iterator iter = attributes.entrySet().iterator(); iter.hasNext();) {
            Map.Entry entry = (Map.Entry) iter.next();
            def key = entry.getKey().toString();
            def val = entry.getValue();
            switch( key) {
                case 'scriptCdata':
                    scriptCdataDefault = Boolean.valueOf(val);
                    scriptCdata = scriptCdataDefault;
                    break;
            }
        }
    }

    def markdown( def value) {
        closePreviousTag()
        try {
            if ( this.mdpClass) {
                out.print( mdpClass.newInstance().markdown( value));
            } else if ( this.pbClass) {
                out.print( pbClass.newInstance().markdown( value));
            }
        } catch( Throwable th) {
            out.print( value);
        }
    }
    
}
